from typing import Callable
import websockets
import json
import requests
from realtime_crypto.models.websocket_details import WebsocketDetails
from realtime_crypto.urls import get_token_data_url


class CryptoWebsocket:
    def __init__(
        self,
        cryptocurrencies: list[str],
        callback: Callable,
    ):
        self.uri = "wss://push.coinmarketcap.com/ws"
        self.headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:129.0) Gecko/20100101 Firefox/129.0"
        }
        self.cryptocurrencies = cryptocurrencies
        self.callback = callback

    async def websocket_init(self):
        crypto_ids = {}
        for name in self.cryptocurrencies:
            res = requests.get(get_token_data_url(name))
            json_res = res.json()
            if "data" in json_res:
                id = str(json_res["data"]["id"])
                crypto_ids[id] = name
            else:
                print(
                    f"{name} not found, please make sure the name is correct in the coinmarketcap url"
                )
        websockets
        async with websockets.connect(self.uri, extra_headers=self.headers) as ws:
            msg = {
                "method": "RSUBSCRIPTION",
                "params": [
                    "main-site@crypto_price_5s@{}@normal",
                    ",".join(list(crypto_ids.keys())),
                ],
            }
            await ws.send(json.dumps(msg))

            # Listen for messages
            async for message in ws:
                msg = json.loads(message)
                if "d" in msg:
                    msg_data = msg["d"]
                    ws_detail = WebsocketDetails(
                        id=msg_data["id"],
                        crypto=crypto_ids[str(msg_data["id"])],
                        price=msg_data["p"],
                        p24=msg_data["p24h"],
                        p7d=msg_data["p7d"],
                        p30d=msg_data["p30d"],
                        p3m=msg_data["p3m"],
                        p1y=msg_data["p1y"],
                        pytd=msg_data["pytd"],
                        p_all=msg_data["pall"],
                        mc=msg_data["mc"],
                    )
                    await self.callback(ws_detail)



# models init
from .coin import Coin
from .crypto_websocket import CryptoWebsocket
from .statistic import Statistic
from .types import Top100Coin, HistoryPoint, GreedFearHistoryPoint, PerformingCrypto
from .websocket_details import WebsocketDetails

__all__ = [
    "Coin",
    "CryptoWebsocket",
    "Statistic",
    "SomeOtherClass",
    "Top100Coin",
    "HistoryPoint",
    "GreedFearHistoryPoint",
    "PerformingCrypto",
    "WebsocketDetails",
]



# API URLS


def get_token_data_url(slug: str | int) -> str:
    return f"https://api.coinmarketcap.com/data-api/v3/cryptocurrency/detail/?{'slug' if isinstance(slug, str) else 'id'}={slug}"


def format_history_url(coin_id: int | str, range: str = "1D") -> str:
    return f"https://api.coinmarketcap.com/data-api/v3/cryptocurrency/detail/chart?id={coin_id}&range={range}"


def get_top_100_url() -> str:
    return "https://api.coingecko.com/api/v3/coins/markets?vs_currency=usd&order=market_cap_desc&per_page=100"


def get_fear_greed_index_url(start: int, end: int) -> str:
    return f"https://api.coinmarketcap.com/data-api/v3/fear-greed/chart?start={start}&end={end}"


def search_coin_url(keyword):
    return {
        "url": "https://api.coinmarketcap.com/gravity/v4/gravity/global-search",
        "body": {"keyword": keyword, "limit": 1, "scene": "community"},
    }



from typing import Optional
import httpx
from realtime_crypto.models.crypto_websocket import CryptoWebsocket
from realtime_crypto.models.types import HistoryPoint
from realtime_crypto.urls import format_history_url, get_token_data_url
from realtime_crypto.models.statistic import Statistic


class Coin:
    def __init__(
        self, id: int, symbol: str, slug: str | int, desc: str, price_stats: Statistic
    ) -> None:
        self.id = id
        self.symbol = symbol
        self.slug = slug
        self.description = desc
        self.price_stats = price_stats
        self.price = price_stats.get_price()

    def get_id(self) -> int:
        return self.id

    def get_slug(self) -> str | int:
        return self.slug

    def get_symbol(self) -> str:
        return self.symbol

    def get_description(self) -> str:
        return self.description

    def get_statistics(self) -> Statistic:
        """
        Get statistics for this coin.
        """
        return self.price_stats

    @classmethod
    def get_current_price(self, slug: str):
        """
        Fetch the current price of a cryptocurrency by its slug.
        Example usage: Coin.get_current_price('bitcoin')
        """

        res = httpx.get(get_token_data_url(slug))
        if res.status_code != 200:
            raise Exception(f"Failed to fetch data for {slug}: {res.text}")

        json_data = res.json()
        if json_data["status"]["error_code"] != "0":
            raise Exception(
                f"Error in response: {json_data['status']['error_message']}"
            )

        new_price = json_data["data"]["statistics"]["price"]
        self.price = new_price
        return new_price

    def get_price(self):
        return self.price

    async def get_realtime_price(self, callback):
        crypto_ws = CryptoWebsocket([self.get_slug()], callback)
        await crypto_ws.websocket_init()

    def get_history(self, range: str) -> Optional[list[HistoryPoint]]:
        """
        Get the price history.
        Example range inputs: 1h, 1d, 7d, 1m, 3m, 1y

        Return example [{'id': '1', 'timestamp': '1723334400', 'price': 60944}]
        """

        history = httpx.get(format_history_url(self.id, range))
        history_json = history.json()

        if history_json["status"]["error_code"] != "0":
            return None

        history_data = history_json["data"]["points"]

        return [
            {"id": self.id, "timestamp": point, "price": history_data[point]["v"][0]}
            for point in history_data
        ]

    def __repr__(self) -> str:
        return f"this is {self.symbol}"



class WebsocketDetails:
    def __init__(
        self,
        id: int,
        crypto: str,
        price: float,
        p24: float,
        p7d: float,
        p30d: float,
        p3m: float,
        p1y: float,
        pytd: float,
        p_all: float,
        mc: float,
    ):
        self.id = id
        self.crypto = crypto
        self.price = price  # current price
        self.p24 = p24  # 24-hour price change
        self.p7d = p7d  # 7-day price change
        self.p30d = p30d  # 30-day price change
        self.p3m = p3m  # 3-month price change
        self.p1y = p1y  # 1-year price change
        self.pytd = pytd  # Year-to-date price change
        self.p_all = p_all  # All-time price change
        self.mc = mc  # Market capitalization

    def get_id(self) -> int:
        return self.id

    def get_new_price(self) -> float:
        return self.price

    def get_crypto(self) -> str:
        return self.crypto

    def get_p24(self) -> float:
        return self.p24

    def get_p7d(self) -> float:
        return self.p7d

    def get_p30d(self) -> float:
        return self.p30d

    def get_p3m(self) -> float:
        return self.p3m

    def get_p1y(self) -> float:
        return self.p1y

    def get_pytd(self) -> float:
        return self.pytd

    def get_p_all(self) -> float:
        return self.p_all

    def get_mc(self) -> float:
        return self.mc



from typing import Callable, Optional
import httpx
from realtime_crypto.models.crypto_websocket import CryptoWebsocket
from realtime_crypto.models.coin import Coin
from realtime_crypto.models.statistic import Statistic
from realtime_crypto.urls import (
    get_token_data_url,
    get_top_100_url,
    format_history_url,
    get_fear_greed_index_url,
    search_coin_url,
)
from realtime_crypto.models.types import (
    PerformingCrypto,
    Top100Coin,
    HistoryPoint,
    GreedFearHistoryPoint,
)
import time
from datetime import datetime, timedelta


class RealTimeCrypto:
    def __init__(self):
        self.client = httpx.Client()

    def _fetch_json(self, url: str) -> Optional[dict]:
        try:
            res = self.client.get(url)
            res.raise_for_status()
            return res.json()
        except httpx.HTTPStatusError as e:
            print("Error occured trying to fetch data")
            return None

    def get_coin(self, coin_name: str | int) -> Optional[Coin]:
        """
        Get data of a single coin.
        """

        res = self._fetch_json(get_token_data_url(coin_name))

        if "data" not in res:
            return None

        json_data = res["data"]

        price_stats = Statistic(**json_data["statistics"])

        id = json_data["id"]
        desc = json_data["description"]
        symbol = json_data["symbol"]

        coin_obj = Coin(id, symbol, coin_name, desc, price_stats)
        return coin_obj

    def get_current_price(self, coin_name: str):
        """
        slug: provide the slug of a coin, for example https://coinmarketcap.com/currencies/bitcoin/ is "bitcoin"
        """

        price_json = self._fetch_json(get_token_data_url(coin_name))

        if price_json["status"]["error_code"] != "0":
            return 0

        price = price_json["data"]["statistics"]["price"]

        return price

    def get_coin_id(self, coin_name: str) -> int:
        """
        Returns the coinmarketcap id of a given crypto
        """

        res_json = self._fetch_json(get_token_data_url(coin_name))

        if "data" not in res_json:
            return 0

        id = res_json["data"]["id"]
        return int(id)

    async def realtime_prices(self, cryptocurrencies: list[str], callback: Callable):
        """
        Joins the coinmarketcap websocket.
        Get realtime prices through the callback function.
        """

        crypto_ws = CryptoWebsocket(cryptocurrencies, callback)
        await crypto_ws.websocket_init()

    def get_top_100_coins(self) -> list[Top100Coin]:
        """
        Get top 100 crypto on coinmarketcap ranked by marketcap.

        Example output
        {
            'name': 'Bitcoin',
            'symbol': 'BTC',
            'price': '58104.2710616952577758'
            'rank': 1
        }

        """
        cryptocurrencies = self._fetch_json(get_top_100_url())

        return [
            {
                "name": crypto["name"],
                "symbol": crypto["symbol"],
                "price": crypto["current_price"],
                "rank": crypto["market_cap_rank"],
            }
            for crypto in cryptocurrencies
        ]

    def get_history(self, coin: int | str, range: str) -> list[HistoryPoint] | None:
        """
        Get the price history for a given.
        Example range inputs: 1h, 1d, 7d, 1m, 3m, 1y

        Return example {'id': '1', 'timestamp': '1723334400', 'price': 60944}
        """

        if isinstance(coin, str):
            coin = self.get_coin_id(coin)
            if not coin:
                return None

        history_json = self._fetch_json(format_history_url(coin, range))

        if history_json["status"]["error_code"] != "0":
            return None

        history_data = history_json["data"]["points"]

        return [
            {"id": coin, "timestamp": point, "price": history_data[point]["v"][0]}
            for point in history_data
        ]

    def get_fear_greed_index_history(
        self, from_unix: int, to_unix: int
    ) -> list[GreedFearHistoryPoint]:
        """
        Get the greed/fear index history for a given time frame.

        Return example:
        [
            {
                "score": 36.62,
                "name": "Fear",
                "timestamp": "1694131200"
            },
            {
                "score": 32.00,
                "name": "Fear",
                "timestamp": "1694931200"
            }
        ]
        """

        res = self._fetch_json(get_fear_greed_index_url(from_unix, to_unix))
        if res is None:
            return []

        res_data = res["data"]
        if "dataList" not in res_data:
            return []

        history_list: list[GreedFearHistoryPoint] = res_data["dataList"]

        return history_list

    def get_current_fear_greed_index(self) -> int:
        """
        Get current fear/greed index.
        """

        # Get yesterday unix
        now = datetime.now()
        yesterday = now - timedelta(days=1)
        yesterday_unix = int(time.mktime(yesterday.timetuple()))

        now_unix = time.time()

        res = self._fetch_json(
            get_fear_greed_index_url(int(yesterday_unix), int(now_unix))
        )
        if res is None:
            return 0

        res_data = res["data"]
        if "dataList" not in res_data:
            return 0

        score = int(res_data["dataList"][0]["score"])

        return score

    def get_best_performing_cryptos(
        self, range: str, reverse=True
    ) -> Optional[list[PerformingCrypto]]:
        """
        Get best performing cryptos in the top 100.

        range examples: 1h, 24h, 7d, 30d, 60d, 90d, 1y

        Return value is sorted high to low
        """

        valid_ranges = ["1h", "24h", "7d", "30d", "60d", "90d", "1y"]
        if range not in valid_ranges:
            return None

        result = []

        top_100_list = self._fetch_json(get_top_100_url())

        if range == "24h":
            sortedli = sorted(
                top_100_list,
                key=lambda d: d["price_change_percentage_24h"],
                reverse=reverse,
            )

            return sortedli

        for coin in top_100_list:
            coin_obj = self.get_coin(coin["id"])

            if coin_obj is None:
                url_data = search_coin_url(coin["name"])
                res = self.client.post(url=url_data["url"], json=url_data["body"])
                searched_coin_slug = res.json()["data"]["suggestions"][1]["tokens"][0][
                    "slug"
                ]
                coin_obj = self.get_coin(searched_coin_slug)
                if coin_obj is None:
                    continue

            price_change = coin_obj.get_statistics().get_price_change(range)
            if price_change:
                result.append(
                    {
                        "symbol": coin_obj.get_symbol(),
                        "slug": coin_obj.get_slug(),
                        "priceChange" + range: price_change,
                    }
                )

        return sorted(result, key=lambda d: d["priceChange" + range], reverse=reverse)

    def search_coin_id(self, search: str) -> Optional[int]:
        """
        Get the coinmarketcap ID if a coin, given a coin name / symbol
        """
        url_data = search_coin_url(search)
        res = self.client.post(url=url_data["url"], json=url_data["body"])
        post_count = res.json()["data"]["suggestions"][0]["postCount"]

        if post_count == "0":
            return None

        searched_coin_id = res.json()["data"]["suggestions"][1]["tokens"][0]["id"]
        return searched_coin_id



# RealTimeCrypto
# Copyright 2024 SanderSita
# See LICENSE for details.

"""
RealTimeCrypto
"""

__version__ = "0.0.1"
__author__ = "SanderSita"
__license__ = "MIT"

from .api import RealTimeCrypto
from .models import (
    Coin,
    CryptoWebsocket,
    Statistic,
    WebsocketDetails,
    Top100Coin,
    HistoryPoint,
    GreedFearHistoryPoint,
    PerformingCrypto,
)



from typing import Optional, TypedDict


class Top100Coin(TypedDict):
    name: str
    symbol: str
    price: float
    rank: int


class HistoryPoint(TypedDict):
    id: int
    timestamp: str
    price: float


class GreedFearHistoryPoint(TypedDict):
    score: float
    name: str
    timestamp: str


class PerformingCrypto(TypedDict):
    id: str
    symbol: str
    name: str
    image: str
    current_price: float
    market_cap: int
    market_cap_rank: int
    fully_diluted_valuation: int
    total_volume: int
    high_24h: float
    low_24h: float
    price_change_24h: float
    price_change_percentage_24h: float
    market_cap_change_24h: int
    market_cap_change_percentage_24h: float
    circulating_supply: float
    total_supply: float
    max_supply: Optional[float]
    ath: float
    ath_change_percentage: float
    ath_date: str
    atl: float
    atl_change_percentage: float
    atl_date: str
    roi: Optional[float]
    last_updated: str



from typing import Optional


class Statistic:
    def __init__(self, **kwargs) -> None:
        self.price = kwargs.get("price")
        self.priceChangePercentage1h = kwargs.get("priceChangePercentage1h")
        self.priceChangePercentage24h = kwargs.get("priceChangePercentage24h")
        self.priceChangePercentage7d = kwargs.get("priceChangePercentage7d")
        self.priceChangePercentage30d = kwargs.get("priceChangePercentage30d")
        self.priceChangePercentage60d = kwargs.get("priceChangePercentage60d")
        self.priceChangePercentage90d = kwargs.get("priceChangePercentage90d")
        self.priceChangePercentage1y = kwargs.get("priceChangePercentage1y")
        self.priceChangePercentageAll = kwargs.get("priceChangePercentageAll")
        self.marketCap = kwargs.get("marketCap")
        self.marketCapChangePercentage24h = kwargs.get("marketCapChangePercentage24h")
        self.fullyDilutedMarketCap = kwargs.get("fullyDilutedMarketCap")
        self.fullyDilutedMarketCapChangePercentage24h = kwargs.get(
            "fullyDilutedMarketCapChangePercentage24h"
        )
        self.circulatingSupply = kwargs.get("circulatingSupply")
        self.totalSupply = kwargs.get("totalSupply")
        self.maxSupply = kwargs.get("maxSupply")
        self.marketCapDominance = kwargs.get("marketCapDominance")
        self.rank = kwargs.get("rank")
        self.roi = kwargs.get("roi")
        self.low24h = kwargs.get("low24h")
        self.high24h = kwargs.get("high24h")
        self.low7d = kwargs.get("low7d")
        self.high7d = kwargs.get("high7d")
        self.low30d = kwargs.get("low30d")
        self.high30d = kwargs.get("high30d")
        self.low90d = kwargs.get("low90d")
        self.high90d = kwargs.get("high90d")
        self.low52w = kwargs.get("low52w")
        self.high52w = kwargs.get("high52w")
        self.lowAllTime = kwargs.get("lowAllTime")
        self.highAllTime = kwargs.get("highAllTime")
        self.lowAllTimeChangePercentage = kwargs.get("lowAllTimeChangePercentage")
        self.highAllTimeChangePercentage = kwargs.get("highAllTimeChangePercentage")
        self.lowAllTimeTimestamp = kwargs.get("lowAllTimeTimestamp")
        self.highAllTimeTimestamp = kwargs.get("highAllTimeTimestamp")
        self.lowYesterday = kwargs.get("lowYesterday")
        self.highYesterday = kwargs.get("highYesterday")
        self.openYesterday = kwargs.get("openYesterday")
        self.closeYesterday = kwargs.get("closeYesterday")
        self.priceChangePercentageYesterday = kwargs.get(
            "priceChangePercentageYesterday"
        )
        self.volumeYesterday = kwargs.get("volumeYesterday")
        self.turnover = kwargs.get("turnover")
        self.ytdPriceChangePercentage = kwargs.get("ytdPriceChangePercentage")
        self.volumeRank = kwargs.get("volumeRank")
        self.volumeMcRank = kwargs.get("volumeMcRank")
        self.mcTotalNum = kwargs.get("mcTotalNum")
        self.volumeTotalNum = kwargs.get("volumeTotalNum")
        self.volumeMcTotalNum = kwargs.get("volumeMcTotalNum")

    def get_price(self) -> Optional[float]:
        return self.price

    def get_price_change(self, range: str) -> Optional[float]:
        """
        Get price change in percentage.

        range examples: 1h, 24h, 7d, 30d, 60d, 90d, 1y
        """

        attribute_name = "priceChangePercentage" + range
        price_change = getattr(self, attribute_name, None)

        return float(price_change) if price_change else None
